@@ -4,7 +4,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2023	PgPool Global Development Group
+ * Copyright (c) 2003-2018	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -145,6 +145,7 @@ pool_start_query(POOL_QUERY_CONTEXT * query_context, char *query, int len, Node
 		query_context->rewritten_query = NULL;
 		query_context->parse_tree = node;
 		query_context->virtual_master_node_id = my_master_node_id;
+		query_context->load_balance_node_id = my_master_node_id;
 		query_context->is_cache_safe = false;
 		query_context->num_original_params = -1;
 		if (pool_config->memory_cache_enabled)
@@ -226,7 +227,7 @@ pool_setall_node_to_be_sent(POOL_QUERY_CONTEXT * query_context)
 			 * In streaming replication mode, if the node is not primary node
 			 * nor load balance node, there's no point to send query.
 			 */
-			if (SL_MODE &&
+			if (SL_MODE && !pool_config->statement_level_load_balance &&
 				i != PRIMARY_NODE_ID && i != sc->load_balance_node_id)
 			{
 				continue;
@@ -358,7 +359,7 @@ pool_virtual_master_db_node_id(void)
 					(errmsg("pool_virtual_master_db_node_id: virtual_master_node_id:%d load_balance_node_id:%d PRIMARY_NODE_ID:%d",
 							node_id, sc->load_balance_node_id, PRIMARY_NODE_ID)));
 
-			if (node_id != sc->load_balance_node_id && node_id != PRIMARY_NODE_ID)
+			if (node_id != sc->query_context->load_balance_node_id && node_id != PRIMARY_NODE_ID)
 			{
 				/*
 				 * Only return the primary node id if we are not processing
@@ -428,16 +429,6 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 	 */
 	pool_clear_node_to_be_sent(query_context);
 
-	/*
-	 * When query match the query patterns in black_query_pattern_list, we
-	 * send only to master node.
-	 */
-	if (MASTER_SLAVE && pattern_compare(query, BLACKLIST, "black_query_pattern_list") == 1)
-	{
-		pool_set_node_to_be_sent(query_context, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID);
-		return;
-	}
-
 	/*
 	 * If there is "NO LOAD BALANCE" comment, we send only to master node.
 	 */
@@ -555,19 +546,6 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 					}
 
-					/*
-					 * If a writing function call is used, we prefer to send
-					 * to the primary.
-					 */
-					else if (pool_has_function_call(node))
-					{
-						ereport(DEBUG1,
-								(errmsg("could not load balance because writing functions are used"),
-								 errdetail("destination = %d for query= \"%s\"", dest, query)));
-
-						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
-					}
-
 					/*
 					 * If system catalog is used in the SELECT, we prefer to
 					 * send to the primary. Example: SELECT * FROM pg_class
@@ -612,11 +590,34 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 
 						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 					}
+					/*
+					 * When query match the query patterns in black_query_pattern_list, we
+					 * send only to master node.
+					 */
+					else if (pattern_compare(query, BLACKLIST, "black_query_pattern_list") == 1)
+					{
+						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
+					}
+					/*
+					 * If a writing function call is used, we prefer to send
+					 * to the primary.
+					 */
+					else if (pool_has_function_call(node))
+					{
+						ereport(DEBUG1,
+								(errmsg("could not load balance because writing functions are used"),
+								 errdetail("destination = %d for query= \"%s\"", dest, query)));
 
+						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
+					}
 					else
 					{
+						if (pool_config->statement_level_load_balance)
+							session_context->load_balance_node_id = select_load_balancing_node();
+
+						session_context->query_context->load_balance_node_id = session_context->load_balance_node_id;
 						pool_set_node_to_be_sent(query_context,
-												 session_context->load_balance_node_id);
+												 session_context->query_context->load_balance_node_id);
 					}
 				}
 				else
@@ -632,10 +633,6 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 			}
 		}
 	}
-	else if (REPLICATION && query_context->is_multi_statement)
-	{
-		pool_setall_node_to_be_sent(query_context);
-	}
 	else if (REPLICATION)
 	{
 		if (pool_config->load_balance_mode &&
@@ -662,9 +659,17 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 					  !pool_is_failed_transaction() &&
 					  pool_get_transaction_isolation() != POOL_SERIALIZABLE))
 			{
+
+
+
 				/* load balance */
+				if (pool_config->statement_level_load_balance)
+					session_context->load_balance_node_id = select_load_balancing_node();
+
+				session_context->query_context->load_balance_node_id = session_context->load_balance_node_id;
+
 				pool_set_node_to_be_sent(query_context,
-										 session_context->load_balance_node_id);
+										 session_context->query_context->load_balance_node_id);
 			}
 			else
 			{
@@ -695,9 +700,24 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 	}
 
 	/*
-	 * DEALLOCATE or EXECUTE?
+	 * EXECUTE?
 	 */
-	if (IsA(node, DeallocateStmt) || IsA(node, ExecuteStmt))
+	if (IsA(node, ExecuteStmt))
+	{
+		POOL_SENT_MESSAGE *msg;
+
+		msg = pool_get_sent_message('Q', ((ExecuteStmt *) node)->name, POOL_SENT_MESSAGE_CREATED);
+		if (!msg)
+			msg = pool_get_sent_message('P', ((ExecuteStmt *) node)->name, POOL_SENT_MESSAGE_CREATED);
+		if (msg)
+			pool_copy_prep_where(msg->query_context->where_to_send,
+								 query_context->where_to_send);
+	}
+
+	/*
+	 * DEALLOCATE?
+	 */
+	else if (IsA(node, DeallocateStmt))
 	{
 		where_to_send_deallocate(query_context, node);
 	}
@@ -855,15 +875,8 @@ pool_send_and_wait(POOL_QUERY_CONTEXT * query_context,
 		 * Check if some error detected.  If so, emit log. This is useful when
 		 * invalid encoding error occurs. In this case, PostgreSQL does not
 		 * report what statement caused that error and make users confused.
-		 * Also set reset_query_error to true in ERROR case. This does
-		 * anything in normal query processing but when processing reset
-		 * queries, this is important because it might mean DISCARD ALL
-		 * command fails. If so, we need to discard the connection cache so
-		 * that any session object (i.e. named statement) does not remain in
-		 * the last session.
 		 */
-		if (per_node_error_log(backend, i, string, "pool_send_and_wait: Error or notice message from backend: ", true) == 'E')
-			reset_query_error = true;
+		per_node_error_log(backend, i, string, "pool_send_and_wait: Error or notice message from backend: ", true);
 	}
 
 	return POOL_CONTINUE;
@@ -1455,87 +1468,35 @@ static POOL_DEST send_to_where(Node *node, char *query)
 	return POOL_PRIMARY;
 }
 
-/*
- * Decide where to send given message.
- * "node" must be a parse tree of either DEALLOCATE or EXECUTE.
- */
 static
 void
 where_to_send_deallocate(POOL_QUERY_CONTEXT * query_context, Node *node)
 {
-	DeallocateStmt *d = NULL;
-	ExecuteStmt *e = NULL;
-	char		*name;
+	DeallocateStmt *d = (DeallocateStmt *) node;
 	POOL_SENT_MESSAGE *msg;
 
-	if (IsA(node, DeallocateStmt))
-	{
-		d = (DeallocateStmt *) node;
-		name = d->name;
-	}
-	else if (IsA(node, ExecuteStmt))
-	{
-		e = (ExecuteStmt *) node;
-		name = e->name;
-	}
-	else
-	{
-		ereport(ERROR,
-				(errmsg("invalid node type for where_to_send_deallocate")));
-		return;
-	}
-
 	/* DEALLOCATE ALL? */
-	if (d && (name == NULL))
+	if (d->name == NULL)
 	{
-		/* send to all backend node */
 		pool_setall_node_to_be_sent(query_context);
-		return;
 	}
-
-	/* ordinary DEALLOCATE or EXECUTE */
 	else
 	{
-		/* check if message was created by SQL PREPARE */
-		msg = pool_get_sent_message('Q', name, POOL_SENT_MESSAGE_CREATED);
+		msg = pool_get_sent_message('Q', d->name, POOL_SENT_MESSAGE_CREATED);
 		if (!msg)
-			/* message may be created by Parse message */
-			msg = pool_get_sent_message('P', name, POOL_SENT_MESSAGE_CREATED);
+			msg = pool_get_sent_message('P', d->name, POOL_SENT_MESSAGE_CREATED);
 		if (msg)
 		{
-			/* Inherit same map from PREPARE or Parse */
+			/* Inherit same map from PREPARE or PARSE */
 			pool_copy_prep_where(msg->query_context->where_to_send,
 								 query_context->where_to_send);
-			return;
+
+			/* copy load balance node id as well */
+			query_context->load_balance_node_id = msg->query_context->load_balance_node_id;
 		}
 		else
-		{
-			/*
-			 * prepared statement was not found.
-			 * There are two cases when this could happen.
-			 * (1) mistakes by client. In this case backend will return ERROR
-			 * anyway.
-			 * (2) previous query was issued as multi-statement query. e.g.
-			 * SELECT 1\;PREPARE foo AS SELECT 1;
-			 * In this case pgpool does not know anything about the prepared
-			 * statement "foo".
-			 */
-			if (SL_MODE)
-			{
-				/*
-				 * In streaming replication or logical replication, sent to
-				 * primary node only.
-				 */
-				pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
-			}
-			else
-			{
-				/*
-				 * In other mode, sent to all node.
-				 */
-				pool_setall_node_to_be_sent(query_context);
-			}
-		}
+			/* prepared statement was not found */
+			pool_setall_node_to_be_sent(query_context);
 	}
 }
 
